#!/usr/bin/env python

# see app/src/cmds/cmd_write_res.cpp

import argparse
import os

parser = argparse.ArgumentParser(description="Embed a list of files into a resource file")
parser.add_argument("-O", "--output", metavar="<file>", required=True, help="set the output code file")
parser.add_argument("-I", "--inc-dir", metavar="<dir>", required=False, help="set the prefix for #include")
parser.add_argument("--inc-as", metavar="<include>", required=False, help="set the argument for #include")
parser.add_argument("-H", "--header", metavar="<file>", required=False, help="set the output header file")
parser.add_argument("--ns", metavar="<namespace>", required=False, help="set the namespace for the resources")
parser.add_argument('input', nargs='+', metavar="<input>", help="set the file to be embedded")
args = parser.parse_args()

if args.inc_dir is None and args.inc_as is None:
    parser.error("at least one of --inc-dir and --inc-as must be prvoided")
if args.inc_dir is not None and args.inc_as is not None:
    parser.error("at most one of --inc-dir and --inc-as must be prvoided")

if args.header is None:
    base, ext = os.path.splitext(args.output)
    ext = {
        '.cc': '.hh',
        '.cpp': '.hpp',
        '.cxx': '.hxx'
    }[ext]
    args.header = base + ext

idents = {}
names = {}
inputs = [os.path.abspath(input) for input in args.input]
prefix = os.path.commonpath(inputs)
for input in inputs:
    ident = os.path.splitext(input[len(prefix):])[0].replace('\\', '/')
    if not len(ident): continue
    if ident[0] == '/': ident = ident[1:]
    # TODO? safe idents out of sub-dir names...
    if '/' in ident: continue
    names[input] = ident
    idents[input] = ident.replace('-', '_')

with open(args.header, "w", encoding="UTF-8") as header:
    header.write("""// THIS FILE IS AUTOGENERATED
#pragma once

#include <cstddef>
#include <string_view>
#include <optional>

// clang-format off
""")

    prefix = ""
    if args.ns is not None:
        header.write(f"""namespace {args.ns} {{
""")
        prefix = "    "

    for input in inputs:
        ident = idents[input]
        header.write(f"""{prefix}struct {ident} {{
{prefix}    static const char* data() noexcept;
{prefix}    static std::size_t size() noexcept;
{prefix}    static std::string_view view() noexcept {{ return {{data(), size()}}; }}
{prefix}}};

""")

    header.write(f"""{prefix}inline std::optional<std::string_view> get_resource(std::string_view key) noexcept {{
#ifdef _MSC_VER
{prefix}    #pragma warning(push)
{prefix}    #pragma warning(disable: 4455)
#endif
{prefix}    using std::literals::operator""sv;
#ifdef _MSC_VER
{prefix}    #pragma warning(pop)
#endif

{prefix}    struct pair {{
{prefix}        std::string_view key;
{prefix}        std::string_view (*get)();
{prefix}    }};

{prefix}    static constexpr pair blobs[] = {{
""")
    for input in inputs:
        ident = idents[input]
        name = names[input]
        header.write(f"""{prefix}        {{"{name}"sv, [] {{ return {ident}::view(); }}}},
""")
    header.write(f"""{prefix}    }};

{prefix}    for (auto const& [blob, get] : blobs) {{
{prefix}        if (blob == key) return get();
{prefix}    }}

{prefix}    return std::nullopt;
{prefix}}};
""")

    if args.ns is not None:
        header.write(f"""}} // namespace {args.ns}
""")
    header.write(f"""// clang-format on
""")

ROW_WIDTH = 16
with open(args.output, "w", encoding="UTF-8") as source:
    include = os.path.relpath(args.header, args.inc_dir) if args.inc_as is None else args.inc_as

    source.write(f"""// THIS FILE IS AUTOGENERATED
#include "{include}"

// clang-format off
""")

    prefix = ""
    if args.ns is not None:
        source.write(f"""namespace {args.ns} {{
""")
        prefix = "    "

    source.write(f"""{prefix}namespace {{
""")

    first = True
    for input in inputs:
        if first: first = False
        else: source.write("\n");
        ident = idents[input]
        source.write(f"""{prefix}    const char {ident}_data[] = {{
""")
        with open(input, "rb") as file:
            content = file.read()
            offset = 0
            for b in content:
                if (offset % ROW_WIDTH) == 0:
                    source.write(f"""{prefix}        \"""");
                source.write(f"\\x{b:02x}")
                if ((offset + 1) % ROW_WIDTH) == 0:
                    source.write("\"\n")
                offset += 1
            if (offset % ROW_WIDTH) != 0:
                source.write("\"\n")

        source.write(f"""{prefix}    }}; // __{ident}
""")

    source.write(f"""{prefix}}} // namespace
""")
    for input in inputs:
        ident = idents[input]
        source.write(f"""
{prefix}/*static*/ const char* {ident}::data() noexcept {{ return {ident}_data; }}
{prefix}/*static*/ std::size_t {ident}::size() noexcept {{ return sizeof({ident}_data) - 1; }}
""")

    if args.ns is not None:
        source.write(f"""}} // namespace {args.ns}
""")
    source.write(f"""// clang-format on
""")
